import redfearn
import unittest


class KnownValues(unittest.TestCase):
    dms_test_values = [
        [redfearn.DMS(-37, 39, 10.15611), -37.652821141666664],
        [redfearn.DMS(143, 55, 35.38393), 143.9264955361111],
        [redfearn.DMS(127, 10, 25.07), 127.17363055555556],
        [redfearn.DMS(-1, 52, 43.22), -1.8786722222222223],
        [redfearn.DMS(125, 17, 41.86), 125.2949611111111],
        [redfearn.DMS(0, 0, -20.67), -0.0057416666666666675],
        [redfearn.DMS(125, 17, 21.18), 125.28921666666666],
        [redfearn.DMS(0, 18, 19.70), 0.3054722222222222],
        [redfearn.DMS(0, 0, 40.14), 0.01115],
        [redfearn.DMS(-37, 57, 03.72030), -37.95103341666667],
        [redfearn.DMS(144, 25, 29.52442), 144.42486789444445],
        [redfearn.DMS(306, 52, 05.37), 306.8681583333333],
        [redfearn.DMS(-1, 35, 03.65), -1.5843472222222224],
        [redfearn.DMS(305, 17, 01.72), 305.28381111111116],
        [redfearn.DMS(0, 0, 19.47), 0.005408333333333333],
        [redfearn.DMS(305, 17, 21.18), 305.2892166666667],
        [redfearn.DMS(0, 0, 0), 0],
        [redfearn.DMS(60, 0, 0), 60],
        [redfearn.DMS(0, 60, 0), 1],
        [redfearn.DMS(0, 0, 60), 0.016666666666666666],
        [redfearn.DMS(-0, 0, 0), 0],
        [redfearn.DMS(-60, 0, 0), -60],
        [redfearn.DMS(0, -60, 0), -1],
        [redfearn.DMS(0, 0, -60), -0.016666666666666666],
    ]

    known_values = [
        {
            'name': 'Buninyong',
            'data': {
                'latitude_dms': redfearn.DMS(-37, 39, 10.15611),
                'longitude_dms': redfearn.DMS(143, 55, 35.38393),
                'ahd': 744.986,
                'n_ausgeoid98': 4.869,
                'ellipsoidal_height': 749.855,
                'coordinate_system': "GDA-MGA",
                'easting': 228854.052,
                'northing': 5828259.038,
                'zone': 55,
                'x': -4087103.458,
                'y': 2977473.0435,
                'z': -3875464.7525,
                'azimuth_dms': redfearn.DMS(127, 10, 25.07),
                'grid_convergence_dms': redfearn.DMS(-1, 52, 43.22),
                'grid_bearing_dms': redfearn.DMS(125, 17, 41.86),
                'arc_to_chord_dms': redfearn.DMS(0, 0, -20.67),
                'plane_bearing_dms': redfearn.DMS(125, 17, 21.18),
                'point_scale_factor': 1.00050567,
                'meridian_distance': -4173410.326,
                'rho': 6359253.8294,
                'nu': 6386118.6742,
                'ellipsoidal_distance': 54972.271,
                'line_scale_factor': 1.00036397,
                'grid_and_plane_distance': 54992.279,
                'meridian_convergence_dms': redfearn.DMS(0, 18, 19.70),
                'line_curvature_dms': redfearn.DMS(0, 0, 40.14)
            }
        },
        {
            'name': 'Buninyong',
            'data': {
                'latitude_dms': redfearn.DMS(-37, 39, 10.15611),
                'longitude_dms': redfearn.DMS(143, 55, 35.38393),
                'ahd': 744.986,
                'n_ausgeoid98': 4.869,
                'ellipsoidal_height': 749.855,
                'coordinate_system': "GDA-MGA",
                'easting': 758173.798008,
                'northing': 5828674.33973,
                'zone': 54,
                'x': -4087103.458,
                'y': 2977473.0435,
                'z': -3875464.7525,
                'azimuth_dms': redfearn.DMS(127, 10, 25.07),
                'grid_convergence_dms': redfearn.DMS(1, 47, 19.360430267999618),
                'grid_bearing_dms': redfearn.DMS(125, 17, 41.86),
                'arc_to_chord_dms': redfearn.DMS(0, 0, -20.67),
                'plane_bearing_dms': redfearn.DMS(125, 17, 21.18),
                'point_scale_factor': 1.00042107306,
                'meridian_distance': -4173410.326,
                'rho': 6359253.8294,
                'nu': 6386118.6742,
                'ellipsoidal_distance': 54972.271,
                'line_scale_factor': 1.00036397,
                'grid_and_plane_distance': 54992.279,
                'meridian_convergence_dms': redfearn.DMS(0, 18, 19.70),
                'line_curvature_dms': redfearn.DMS(0, 0, 40.14)
            }
        },
        {
            'name': 'Flinders Peak',
            'data': {
                'latitude_dms': redfearn.DMS(-37, 57, 03.72030),
                'longitude_dms': redfearn.DMS(144, 25, 29.52442),
                'ahd': 347.200,
                'n_ausgeoid98': 3.748,
                'ellipsoidal_height': 350.948,
                'coordinate_system': "GDA-MGA",
                'easting': 273741.297,
                'northing': 5796489.777,
                'zone': 55,
                'x': -4096088.424,
                'y': 2929823.08435,
                'z': -3901375.4540,
                'azimuth_dms': redfearn.DMS(306, 52, 05.37),
                'grid_convergence_dms': redfearn.DMS(-1, 35, 03.65),
                'grid_bearing_dms': redfearn.DMS(305, 17, 01.72),
                'arc_to_chord_dms': redfearn.DMS(0, 0, 19.47),
                'plane_bearing_dms': redfearn.DMS(305, 17, 21.18),
                'point_scale_factor': 1.00023056,
                'meridian_distance': -4205192.300,
                'rho': 6359576.5731,
                'nu': 6386226.7080,
                'ellipsoidal_distance': 54972.271,
                'line_scale_factor': 1.00036397,
                'grid_and_plane_distance': 54992.279,
                'meridian_convergence_dms': redfearn.DMS(0, 18, 19.70),
                'line_curvature_dms': redfearn.DMS(0, 0, 40.14)
            }
        }
    ]

    # def test_decdeg2dms(self):
    #    """ Test the decimal degree to dms conversion
    #    """
    #    for i in self.dms_test_values:
    #        self.assertEqual(redfearn.decdeg2dms(i[1]), i[0])

    def test_dms2decdeg(self):
        """ Test the dms to decimal degrees conversion
        """
        for i in self.dms_test_values:
            self.assertEqual(redfearn.dms2decdeg(i[0]), i[1])

    def test_dms2decdeg_and_decdeg2dms_cycle(self):
        """ Test the dms to decimal conversion by cycling between the two.

            dms2decdeg and decdeg2dms should cycle accurately between values without variation
            there is occasionally variation between the initial dms seconds value and the second
            cycle due to floating point errors but the decimal expansion should be identical to at least 12 decimal places

            There can also be variations due to different valid dms values, e.g. 0 deg, 60 min, 0 sec is 1.0 degrees
            or could also be written as 1 deg, 0 min, 0 sec.
        """
        for i in self.dms_test_values:
            dd1 = redfearn.dms2decdeg(i[0])
            dms2 = redfearn.decdeg2dms(dd1)
            dd2 = redfearn.dms2decdeg(dms2)
            self.assertAlmostEqual(dd1, dd2, 12)

    def test_latlon2grid_conversion(self):
        """ Test the latlon2grid conversion.

            eastings and northings should be within 1 millimetre of known values
            grid_convergence is tested to 5 decimal places
            point_scale_factor is tested to 8 decimal places
        """
        for known_site in self.known_values:
            latitude_dd = redfearn.dms2decdeg(known_site['data']['latitude_dms'])
            longitude_dd = redfearn.dms2decdeg(known_site['data']['longitude_dms'])
            grid = redfearn.latlon2grid(latitude_dd, longitude_dd, known_site['data']['coordinate_system'], zone=known_site['data']['zone'])
            self.assertAlmostEqual(grid['easting'], known_site['data']['easting'], delta=0.001)
            self.assertAlmostEqual(grid['northing'], known_site['data']['northing'], delta=0.001)
            self.assertEqual(grid['zone'], known_site['data']['zone'])
            self.assertAlmostEqual(grid['grid_convergence'], redfearn.dms2decdeg(known_site['data']['grid_convergence_dms']), 5)
            self.assertAlmostEqual(grid['point_scale_factor'], known_site['data']['point_scale_factor'], 8)

    def test_grid2latlon_conversion(self):
        """ Test the grid2latlon conversion.

            eastings and northings should be within 1 millimetre of known values
            grid_convergence is tested to 5 decimal places
            point_scale_factor is tested to 8 decimal places
        """
        for known_site in self.known_values:
            latlon = redfearn.grid2latlon(
                known_site['data']['easting'],
                known_site['data']['northing'],
                known_site['data']['zone'],
                coordinate_system=known_site['data']['coordinate_system']
            )
            self.assertAlmostEqual(latlon['latitude'], redfearn.dms2decdeg(known_site['data']['latitude_dms']), 8)
            self.assertAlmostEqual(latlon['longitude'], redfearn.dms2decdeg(known_site['data']['longitude_dms']), 8)
            self.assertAlmostEqual(latlon['grid_convergence'], redfearn.dms2decdeg(known_site['data']['grid_convergence_dms']), 5)
            self.assertAlmostEqual(latlon['point_scale_factor'], known_site['data']['point_scale_factor'], 8)

    def test_latlon2grid_and_grid2latlon_cycle(self):
        """ Test that converting from lat/lon to grid and back to lat/lon occurs correctly.

            eastings and northings should be within 1 millimetre of known values
            grid_convergence is tested to 5 decimal places
            point_scale_factor is tested to 8 decimal places
        """
        for known_site in self.known_values:
            latitude_dd = redfearn.dms2decdeg(known_site['data']['latitude_dms'])
            longitude_dd = redfearn.dms2decdeg(known_site['data']['longitude_dms'])
            grid = redfearn.latlon2grid(latitude_dd, longitude_dd, 'GDA-MGA', zone=known_site['data']['zone'])
            latlon = redfearn.grid2latlon(
                grid['easting'],
                grid['northing'],
                grid['zone'],
                coordinate_system=known_site['data']['coordinate_system']
            )
            self.assertAlmostEqual(latlon['latitude'], redfearn.dms2decdeg(known_site['data']['latitude_dms']), 8)
            self.assertAlmostEqual(latlon['longitude'], redfearn.dms2decdeg(known_site['data']['longitude_dms']), 8)
            self.assertAlmostEqual(latlon['grid_convergence'], redfearn.dms2decdeg(known_site['data']['grid_convergence_dms']), 5)
            self.assertAlmostEqual(latlon['point_scale_factor'], known_site['data']['point_scale_factor'], 8)

    def test_grid2latlon_and_latlon2grid_cycle(self):
        """ Test that converting from grid to lat/lon and back to grid occurs correctly.

            eastings and northings should be within 1 millimetre of known values
            grid_convergence is tested to 5 decimal places
            point_scale_factor is tested to 8 decimal places
        """
        for known_site in self.known_values:
            latlon = redfearn.grid2latlon(
                known_site['data']['easting'],
                known_site['data']['northing'],
                known_site['data']['zone'],
                coordinate_system=known_site['data']['coordinate_system']
            )
            grid = redfearn.latlon2grid(latlon['latitude'], latlon['longitude'], known_site['data']['coordinate_system'], zone=known_site['data']['zone'])
            self.assertAlmostEqual(grid['easting'], known_site['data']['easting'], delta=0.001)
            self.assertAlmostEqual(grid['northing'], known_site['data']['northing'], delta=0.001)
            self.assertEqual(grid['zone'], known_site['data']['zone'])
            self.assertAlmostEqual(grid['grid_convergence'], redfearn.dms2decdeg(known_site['data']['grid_convergence_dms']), 5)
            self.assertAlmostEqual(grid['point_scale_factor'], known_site['data']['point_scale_factor'], 8)

if __name__ == '__main__':
    unittest.main()


from __future__ import division
import math
from collections import namedtuple


DMS = namedtuple('DMS', 'degrees minutes seconds')


def merge_two_dicts(x, y):
    """ Given two dicts, merge them into a new dict as a shallow copy."""
    z = x.copy()
    z.update(y)
    return z


def decdeg2dms(dd):
    """ Converts a decimal degree value of the for -37.4568 into a DMS namedtuple type with degrees, minutes, seconds.
    :param dd: float
    :return: namedtuple DMS
    """
    negative = dd < 0
    dd = abs(dd)
    minutes, seconds = divmod(dd*3600, 60)
    degrees, minutes = divmod(minutes, 60)
    if negative:
        if degrees > 0:
            degrees = -degrees
        elif minutes > 0:
            minutes = -minutes
        else:
            seconds = -seconds
    return DMS(degrees, minutes, seconds)


def dms2decdeg(dms):
    """ Converts the DMS namedtuple into a decimal expansion of the degrees.
    :param dms: namedtuple DMS
    :return: float
    """
    dd = abs(dms.degrees) + abs(dms.minutes)/60 + abs(dms.seconds)/3600
    if dms.degrees < 0 or dms.minutes < 0 or dms.seconds < 0:
        dd = -dd
    return dd


def ellipsoid_constants(name):
    """ Gets known ellipsoid constants vor various coordinate systems as defined in "Geocentric Datum of Australia, Technical Manual, Version 2.3 Amendment 1".
    :param name: string
    :return: dict
    """
    if name == 'GRS80':
        semi_major_axis = 6378137.000
        inverse_flattening = 298.257222101000
    elif name == 'WGS84':
        semi_major_axis = 6378137.000
        inverse_flattening = 298.257223563000
    elif name == 'ANS':
        semi_major_axis = 6378160.000
        inverse_flattening = 298.25
    elif name == 'Clarke 1858':
        semi_major_axis = 6378293.645
        inverse_flattening = 294.26
    else:
        raise ValueError('invalid ellipsoid name: "{}"'.format(name))
    return {'semi_major_axis': semi_major_axis, 'inverse_flattening': inverse_flattening}


def tm_constants(name):
    """ Gets known transverse mercator projection constants vor various coordinate systems as defined in "Geocentric Datum of Australia, Technical Manual, Version 2.3 Amendment 1".
    :param name: string
    :return: dict
    """
    if name == 'GDA-MGA':
        false_easting = 500000.0000
        false_northing = 10000000.0000
        central_scale_factor = 0.9996
        zone_width_degrees = 6
        longitude_of_the_central_meridian_of_zone_1_degrees = -177
    elif name == 'ANG':
        false_easting = 300000.0000
        false_northing = 5000000.0000
        central_scale_factor = 0.99994
        zone_width_degrees = 2
        longitude_of_the_central_meridian_of_zone_1_degrees = 141
    elif name == 'ISG':
        false_easting = 365760.0000
        false_northing = 4495019.83
        central_scale_factor = 1.0
        zone_width_degrees = 5
        longitude_of_the_central_meridian_of_zone_1_degrees = 116
    else:
        raise ValueError('invalid transverse mercator name: "{}"'.format(name))
    return {
        'false_easting': false_easting,
        'false_northing': false_northing,
        'central_scale_factor': central_scale_factor,
        'zone_width_degrees': zone_width_degrees,
        'longitude_of_the_central_meridian_of_zone_1_degrees': longitude_of_the_central_meridian_of_zone_1_degrees
    }


def coordinate_system_constants(name):
    """ Collates the known ellipsoid and transverse mercator projection constants for various coordinate systems
    :param name:
    :return: dict
    """
    if name == 'GDA/MGA' or name == 'GDA-MGA' or name == 'GDA' or name == 'MGA':
        ellipsoid = 'GRS80'
        tm = 'GDA-MGA'
    elif name == 'WGS84':
        ellipsoid = 'WGS84'
        tm = 'GDA-MGA'
    elif name == 'AGD/AMG' or name == 'AGD' or name == 'AMG':
        ellipsoid = 'ANS'
        tm = 'GDA-MGA'
    elif name == 'ANG':
        ellipsoid = 'Clarke 1858'
        tm = 'ANG'
    elif name == 'ISG':
        ellipsoid = 'ANS'
        tm = 'ISG'
    else:
        raise ValueError('unknown coordinate system: "{}"'.format(name))

    e_constants = ellipsoid_constants(ellipsoid)
    t_constants = tm_constants(tm)
    return merge_two_dicts(e_constants, t_constants)


def check_latlon_type(val):
    """ Checks whether a given latitude or longitude is a float or a DMS type and returns the decimal value
    :param val: float or DMS
    :return: float
    """
    try:
        r = float(val)
    except TypeError:
        try:
            r = dms2decdeg(val)
        except ValueError:
            raise
    return r


# latitude and longitude should be in degrees
def latlon2grid(latitude, longitude, coordinate_system=None, zone=None):
    """ Convert latitude and longitude to grid coordinates using the Redfearn Formula
    :param latitude: float (decimal degrees i.e. from dms2decdeg(dms))
    :param longitude: float (decimal degrees i.e. from dms2decdeg(dms))
    :param coordinate_system: str
    :param zone: int
    :return: dict
    """
    if coordinate_system is None:
        coordinate_system = 'GDA-MGA'
    constants = coordinate_system_constants(coordinate_system)

    # check whether latitude and longitude are in decimal form
    try:
        latitude = check_latlon_type(latitude)
        longitude = check_latlon_type(longitude)
    except ValueError:
        raise ValueError('latitude and longitude should be decimal degrees or DMS types as returned by dms2decdeg and decdeg2dms')
    # refer to Geocentric Datum of Australia, Technical Manual, Version 2.3 Amendment 1
    # pages 18 - 21

    # latitude in radians
    latitude_rad = math.radians(latitude)
    s = math.sin(latitude_rad)
    s2 = math.pow(s, 2)
    c = math.cos(latitude_rad)
    c3 = math.pow(c, 3)
    c5 = math.pow(c, 5)
    c7 = math.pow(c, 7)
    t = math.tan(latitude_rad)
    t2 = math.pow(t, 2)
    t4 = math.pow(t, 4)
    t6 = math.pow(t, 6)
    a = constants['semi_major_axis']
    i_f = constants['inverse_flattening']
    # flattening
    f = 1/i_f
    # semi-minor axis
    b = a*(1-f)
    # eccentricity squared, ^4, ^6
    e2 = (math.pow(a, 2) - math.pow(b, 2))/math.pow(a, 2)
    e4 = math.pow(e2, 2)
    e6 = math.pow(e2, 3)

    longitude_of_western_edge_of_zone_zero_degrees = constants['longitude_of_the_central_meridian_of_zone_1_degrees'] - (1.5 * constants['zone_width_degrees'])
    central_meridian_of_zone_zero_degrees = longitude_of_western_edge_of_zone_zero_degrees + (constants['zone_width_degrees'] / 2)
    zone_no_real = (longitude - longitude_of_western_edge_of_zone_zero_degrees) / constants['zone_width_degrees']
    if zone is None:
        zone = math.floor(zone_no_real)
    central_meridian = (zone * constants['zone_width_degrees']) + central_meridian_of_zone_zero_degrees

    # calculate meridian distance
    a0 = 1 - (e2/4) - (3*e4/64) - (5*e6/256)
    a2 = (3/8)*(e2 + e4/4 + 15*e6/128)
    a4 = (15/256)*(e4 + 3*e6/4)
    a6 = 35*e6/3072
    # meridian distance
    m = a*(a0*latitude_rad - a2*math.sin(2*latitude_rad) + a4*math.sin(4*latitude_rad) - a6*math.sin(6*latitude_rad))

    # calculate radius of curvature
    rho = a*(1-e2) / math.pow(1-e2*s2, 3/2)
    nu = a / math.sqrt(1-e2*s2)
    # radius of curvature
    psi = nu / rho
    psi2 = math.pow(psi, 2)
    psi3 = math.pow(psi, 3)
    psi4 = math.pow(psi, 4)
    omega = math.radians(longitude - central_meridian)
    omega2 = math.pow(omega, 2)
    omega4 = math.pow(omega, 4)
    omega6 = math.pow(omega, 6)
    omega8 = math.pow(omega, 8)

    # Easting
    easting_term1 = (omega2/6)*c3*(psi-t2)
    easting_term2 = (omega4/120) * c5 * (4*psi3*(1-6*t2) + psi2*(1+8*t2) - psi*2*t2 + t4)
    easting_term3 = (omega6/5040)*c7*(61-479*t2+179*t4-t6)
    eprime = (constants['central_scale_factor'] * nu * omega) * (c + easting_term1 + easting_term2 + easting_term3)
    easting = eprime + constants['false_easting']

    # Northing
    northing_term1 = (omega2/2)*nu*s*c
    northing_term2 = (omega4/24)*nu*s*c3*(4*psi2 + psi - t2)
    northing_term3 = (omega6/720)*nu*s*c5*(8*psi4*(11-24*t2) - 28*psi3*(1-6*t2) + psi2*(1-32*t2) - psi*(2*t2) + t4)
    northing_term4 = (omega8/40320)*nu*s*c7*(1385 - 3111*t2 + 543*t4 - t6)
    true_northing = constants['central_scale_factor']*(m + northing_term1 + northing_term2 + northing_term3 + northing_term4)
    northing = true_northing + constants['false_northing']

    # grid convergence
    gc_term1 = 1
    gc_term2 = (omega2/3)*c3*(2*psi2 - psi)
    gc_term3 = (omega4/15)*c5*(psi4*(11-24*t2) - psi3*(11-36*t2) + 2*psi2*(1-7*t2) + psi*t2)
    gc_term4 = (omega6/315)*c7*(17 - 26*t2 + 2*t4)
    gc = math.degrees((-omega*s)*(gc_term1 + (1/c)*(gc_term2 + gc_term3 + gc_term4)))

    # point scale factor
    psf_term1 = (omega2/2)*psi*c
    psf_term2 = (omega4/24)*c3*(4*psi3*(1-6*t2) + psi2*(1+24*t2) - 4*psi*t2)
    psf_term3 = (omega6/720)*c5*(61 - 148*t2 + 16*t4)
    psf = constants['central_scale_factor']*(1 + c*(psf_term1 + psf_term2 + psf_term3))

    return {
        'easting': easting,
        'northing': northing,
        'zone': zone,
        'grid_convergence': gc,
        'point_scale_factor': psf
    }


def grid2latlon(easting, northing, zone, coordinate_system=None):
    """ Converts grid coordinates to latitude and longitude values using the Redfearn Formula
    :param easting: float
    :param northing: float
    :param zone: int
    :param coordinate_system: string
    :return: dict
    """
    if coordinate_system is None:
        coordinate_system = 'GDA-MGA'
    constants = coordinate_system_constants(coordinate_system)
    # refer to Geocentric Datum of Australia, Technical Manual, Version 2.3 Amendment 1
    # pages 18 - 21

    # semi-major axis
    a = constants['semi_major_axis']
    # inverse flattening
    i_f = constants['inverse_flattening']
    # flattening
    f = 1/i_f
    # semi-minor axis
    b = a*(1-f)
    # eccentricity squared, ^4, ^6
    e2 = (math.pow(a, 2) - math.pow(b, 2))/math.pow(a, 2)

    eprime = easting - constants['false_easting']
    eprime_scaled = eprime/constants['central_scale_factor']
    nprime = northing - constants['false_northing']
    nprime_scaled = nprime/constants['central_scale_factor']

    # Calculate Foot-point latitude
    n = (a-b)/(a+b)
    n2 = math.pow(n, 2)
    n3 = math.pow(n, 3)
    n4 = math.pow(n, 4)
    g = a * (1-n) * (1-n2) * (1 + (9/4)*n2 + (225/64)*n4) * (math.pi/180)
    sigma = math.radians(nprime_scaled / g)
    fpl = sigma + ((3*n/2) - (27*n3/32))*math.sin(2*sigma) + ((21*n2/16) - (55*n4/32))*math.sin(4*sigma) + (151*n3/96)*math.sin(6*sigma) + (1097*n4/512)*math.sin(8*sigma)
    s = math.sin(fpl)
    s2 = math.pow(s, 2)
    sec = 1 / math.cos(fpl)
    t = math.tan(fpl)
    t2 = math.pow(t, 2)
    t4 = math.pow(t, 4)
    t6 = math.pow(t, 6)

    # calculate radius of curvature
    rho = a*(1-e2) / math.pow(1-e2*s2, 3/2)
    nu = a / math.sqrt(1-e2*s2)
    x = eprime_scaled / nu
    x3 = math.pow(x, 3)
    x5 = math.pow(x, 5)
    x7 = math.pow(x, 7)
    # radius of curvature
    psi = nu / rho
    psi2 = math.pow(psi, 2)
    psi3 = math.pow(psi, 3)
    psi4 = math.pow(psi, 4)

    # latitude
    latitude_term1 = -((t/(constants['central_scale_factor']*rho))*x*eprime/2)
    latitude_term2 = (t/(constants['central_scale_factor'] * rho)) * (x3 * eprime / 24) * (-4 * psi2 + 9 * psi * (1 - t2) + 12 * t2)
    latitude_term3 = -(t/(constants['central_scale_factor']*rho))*(x5*eprime/720)*(8*psi4*(11-24*t2)-12*psi3*(21-71*t2)+15*psi2*(15-98*t2+15*t4)+180*psi*(5*t2-3*t4)+360*t4)
    latitude_term4 = (t/(constants['central_scale_factor']*rho)) * (x7*eprime/40320) * (1385 + 3633*t2 + 4095*t4 + 1575*t6)
    latitude = math.degrees(fpl + latitude_term1 + latitude_term2 + latitude_term3 + latitude_term4)

    # longitude
    central_meridian_deg = (zone*constants['zone_width_degrees']) + constants['longitude_of_the_central_meridian_of_zone_1_degrees'] - constants['zone_width_degrees']
    central_meridian_rad = math.radians(central_meridian_deg)
    longitude_term1 = sec*x
    longitude_term2 = -sec*(x3/6) * (psi + 2*t2)
    longitude_term3 = sec*(x5/120) * (-4*psi3 * (1 - 6*t2) + psi2*(9 - 68*t2) + 72*psi*t2 + 24*t4)
    longitude_term4 = -sec*(x7/5040) * (61 + 662*t2 + 1320*t4 + 720*t6)
    longitude = math.degrees(central_meridian_rad + longitude_term1 + longitude_term2 + longitude_term3 + longitude_term4)

    # grid convergence
    gc_term1 = -x*t
    gc_term2 = (t*x3/3) * (-2*psi2 + 3*psi + t2)
    gc_term3 = -(t*x5/15) * (psi4*(11 - 24*t2) - 3*psi3*(8 - 23*t2) + 5*psi2*(3 - 14*t2) + 30*psi*t2 + 3*t4)
    gc_term4 = (t*x7/315) * (17 + 77*t2 + 105*t4 + 45*t6)
    grid_convergence = math.degrees(gc_term1 + gc_term2 + gc_term3 + gc_term4)

    # point scale factor
    pc_factor1 = pow(eprime_scaled, 2) / (rho*nu)
    pc_factor2 = pow(pc_factor1, 2)
    pc_factor3 = pow(pc_factor1, 3)
    pc_term1 = 1 + pc_factor1/2
    pc_term2 = (pc_factor2/24) * (4 * psi * (1 - 6*t2) - 3*(1 - 16*t2) - 24*t2/psi)
    pc_term3 = pc_factor3/720
    point_scale = constants['central_scale_factor'] * (pc_term1 + pc_term2 + pc_term3)

    return {
        'latitude': latitude,
        'longitude': longitude,
        'grid_convergence': grid_convergence,
        'point_scale_factor': point_scale
    }


from setuptools import setup

setup(
    name='redfearn',
    version='1.0.1',
    description="Converts between latitude/longitude and grid coordinates using Redfearn's Formula",
    author='PHil Howarth',
    author_email='phil@plaintech.net.au',
    url='http://plaintech.net.au/redfearn',
    long_description=(
        "A python module to use Redfearn's formula to convert between latitude/longitude pairs and various grid coordinate systems used in Australia."
        "Based on the document titled 'Geocentric Datum of Australia - Technical Manual Version 2.3 Amendment 1' "
        "produced by the 'Intergovernmental Committee on Surveying and Mapping', ISBN 0-9579951-0-5"
    ),
    keywords=["conversion", "GDA", "MGA", "latitude", "longitude"],
    license="GPLv3+",
    classifiers=[
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Development Status :: 4 - Beta",
        "Environment :: Other Environment",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Scientific/Engineering :: GIS",
        ],
    py_modules=['redfearn', 'tests']
)
